#include <time.h>
#include <stdio.h>
#include <string.h>
#include "tcl4mirc.h"

/* Message loop data structures */
#define TCL4MIRC_MSGLOOP (WM_USER+900)

static WNDPROC mWndProc;
static HWND mHwnd;
static HANDLE msgLoopThread;

/* Initialization function prototypes */
static Tcl_Interp *globalInterp;

/* mIRC communication */

int mIRC_Eval(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
	Tcl_Obj *command = Tcl_NewStringObj("", 0), *space = Tcl_NewStringObj(" ", 1);
	Tcl_Obj *result = NULL;
	char *sresult;
	int len, evalTimes, obji = 1;

	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "?evalTimes? command");
		return TCL_ERROR; 
	}
	
	/* If first argument is numeric, then it is the evalTimes parameter */
	if (Tcl_GetIntFromObj(interp, objv[obji], &evalTimes) == TCL_ERROR) {
		evalTimes = CLB_DEFAULT_EVALTIMES; 
		obji--;
	} 

	/* Join list of strings into command */
	for (obji++; obji < objc; obji++) {
		Tcl_AppendObjToObj(command, objv[obji]);
		if (obji < objc - 1) {
			Tcl_AppendObjToObj(command, space);
		}
	}

	/* Get the result into a TCL object */
	sresult = CLBExecute2((const char *)Tcl_GetStringFromObj(command, &len), evalTimes);
	if (sresult) {
		result = Tcl_NewStringObj(sresult, -1);
		free(sresult);

		Tcl_SetObjResult(interp, result);
	}

	return TCL_OK;
}

int Tcl4mIRC_Eval(Tcl_Interp *interp, char *data, BOOL show) {
	Tcl_Channel tclStderr;
	if (Tcl_EvalEx(interp, data, -1, TCL_EVAL_GLOBAL) == TCL_ERROR) {
		if (show) {
			if (tclStderr = Tcl_GetStdChannel(TCL_STDERR)) {
				Tcl_WriteObj(tclStderr, Tcl_GetObjResult(interp));
			}
		}
		return TCL_ERROR;
	}
	else {
		strncpy(data, Tcl_GetStringResult(interp), MIRC_MAXSTRING);
	}

	return TCL_OK;
}

/* Initialization methods */

static int Tcl4mIRC_GetLoadPath(Tcl_Interp *interp) {
	int evalResult, pathlen = 0;
	char *loadpath, *setpath;

	loadpath = CLBGetLoadPath("echo puts $auto_path | tclsh");

	if (loadpath && *loadpath != 0) {
		/* Build the set auto_path "PATH" command and evaluate it" */
		pathlen = (int)strlen(loadpath);
		*(loadpath+pathlen-2) = 0; /* Remove \r\n at the end of path info */
		setpath = malloc(pathlen + 62);
		snprintf(setpath, pathlen + 62, "set auto_path {%s}; set tcl_library [lindex $auto_path 0]", loadpath);

		evalResult = Tcl_EvalEx(interp, setpath, -1, TCL_EVAL_GLOBAL);

		free(setpath);

		return evalResult;
	}
	return TCL_ERROR;
}

static int Tcl4mIRC_Init(Tcl_Interp *interp) {
	int hasLoadPath;

	Tcl_FindExecutable(MDLL_NAME);
	hasLoadPath = Tcl4mIRC_GetLoadPath(interp);

	/* Set argv and argc because some scripts rely on it */
	Tcl_SetVar(interp, "argc", "0", TCL_GLOBAL_ONLY);
	Tcl_SetVar(interp, "argv",  "", TCL_GLOBAL_ONLY);
	Tcl_SetVar(interp, "argv0", "", TCL_GLOBAL_ONLY);

	Tcl_Init(interp);
	Io_Init(interp);
	Embed_Init(interp);

	/* Now initialize our functions */
	Tcl_CreateObjCommand(interp, "mirc", mIRC_Eval, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_PkgProvide(interp, "mirc", MDLL_VERSION);

	if (hasLoadPath == TCL_ERROR) {
		Tcl_WriteChars(Tcl_GetStdChannel(TCL_STDERR), 
			"You don't have Tcl installed or tclsh.exe is not in your PATH, Tcl's $auto_path cannot be set.", 
			-1);
	}

	return TCL_OK;
}

static Tcl_Interp *Tcl4mIRC_CreateInterp() {
	Tcl_Interp *interp = NULL;

#ifdef USE_TCL_STUBS /* Dynamically link to tcl.dll */
	HMODULE hTcl = NULL;
	TCHAR szLibrary[16];
	int nMinor;
	Tcl_Interp *(*createInterpFunc)();

     for (nMinor = 5; nMinor > 2; nMinor--) {
         wsprintf(szLibrary, _T("tcl8%d.dll"), nMinor);
         hTcl = LoadLibrary(szLibrary);

		 if (hTcl) {
			 createInterpFunc = (Tcl_Interp *(*)())GetProcAddress(hTcl, "Tcl_CreateInterp");
			 interp = createInterpFunc();
			 Tcl_InitStubs(interp, TCL_VERSION, 1);
			 break;
		 }
     }
#else
	interp = Tcl_CreateInterp();
#endif

	return interp;
}

static LRESULT CALLBACK Tcl4mIRC_MessageLoopProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) { 
	if (uMsg == TCL4MIRC_MSGLOOP) { /* Run the Tcl event loop */
			
		/* Handle all events on the Tcl event queue */
		while (globalInterp && Tcl_DoOneEvent(TCL_DONT_WAIT | TCL_ALL_EVENTS));

		return 0;
	} 

	/* Send the rest to mIRC */
	return CallWindowProc(mWndProc, hwnd, uMsg, wParam, lParam);
} 

static DWORD WINAPI Tcl4mIRC_MessageLoopInit(LPVOID parameter) {
	while (globalInterp) {
		PostMessage(mHwnd, TCL4MIRC_MSGLOOP, 0, 0);
		Sleep(1);
	}

	return 0;
}

/* mIRC exported DLL functions */

int __declspec(dllexport) __stdcall tcl_eval(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause) {
	if (CLBInitialize(&data) == FALSE) return 2;

	if (Tcl4mIRC_Eval(globalInterp, data, show) == TCL_ERROR) {
		CLB_HALT(data);
	}

	return 3;
}

/* Disable the WNDPROC pointer truncation warnings */
#pragma warning(push)
#pragma warning(disable:4311)
#pragma warning(disable:4312)

void __declspec(dllexport) __stdcall LoadDll(LOADINFO *t) {
	CLBStartup(t, MDLL_SIG);

	/* Create interpreter using stubs or static method */
	globalInterp = Tcl4mIRC_CreateInterp();
	Tcl4mIRC_Init(globalInterp);

	/* Setup window subclass to hook the Tcl event loop */
	mHwnd = t->mHwnd;
	mWndProc = (WNDPROC)SetWindowLong(mHwnd, GWL_WNDPROC, (LONG)Tcl4mIRC_MessageLoopProc); 

	/* Start the message loop */
	msgLoopThread = CreateThread(NULL, 0, Tcl4mIRC_MessageLoopInit, NULL, 0, NULL);
}

int  __declspec(dllexport) __stdcall UnloadDll(int mTimeout) {
    if (mTimeout == 0) { /* user called /dll -u (or mIRC is shutting down) */
		/* Remove window hook */
		SetWindowLong(mHwnd, GWL_WNDPROC, (LONG)mWndProc);

		/* Unregister standard channels */
		Io_Shutdown(globalInterp);

		/* Delete interpreter */
		Tcl_DeleteInterp(globalInterp);
		globalInterp = NULL;

		/* Kill loop thread */
		if (WaitForSingleObject(msgLoopThread, 1000) == WAIT_TIMEOUT) {
			TerminateThread(msgLoopThread, 1);
		}

		CLBShutdown();
    }
	return 0;
}

#pragma warning(pop)
